(***********************************************************************)
(*                                  DEF.M                              *)
(* Bernd Fiedler, Fakultaet Math/Inf, Universitaet Leipzig, 05.09.1996 *)
(*                                                                     *)
(*       Mathematica-Quelltext der Definitionen zur Symbolischen       *)
(*       Vektoranalysis im Rahmen des Paketes VECTAN Version 1.1       *)
(*                                                                     *)
(***********************************************************************)

(****************************** DefineVector ***************************)
(*
Ein Vektorfeld v hat in VECTAN die Form Vect[v]. Damit Vect[v] durch
seinen Namen v eingegeben werden kann, wird zu dem Symbol v von der
Prozedur DefineVector eine Definition

                        v = Vect[v]

erzeugt. Hierzu muss fuer Vect[ ] das Attribut HoldFirst gesetzt
werden, da sich andernfalls eine Endlosschleife bei Aufruf von v
ergeben wuerde.
*)

Attributes[Vect] = {HoldFirst}


DefineVector[v_List] := (DefineVector /@ v ;)

DefineVector[v_] := 
(
 Off[General::stop];
 If[VectQ[v],
   (* then *)
   Message[DefineVector::error,v],
   (* else *)
   v = Vect[v];
   Message[DefineVector::inf,v]
   ];
 On[General::stop];
)


VectQ[v_] :=
(*
Uebergibt True, falls der Ausdruck v einen der Koepfe
              Vect, grad, rot, VP
besitzt, andernfalls False.
*)
If[
  Or[
    Head[v] === Vect,
    Head[v] === grad,
    Head[v] === rot,
    Head[v] === VP
    ],
  (* then *)
  True,
  (* else *)
  False
  ]



(******************************* Formatting ****************************)
(*
Bei der Ausgabe von Vect[v] soll nur v erscheinen.
*)

Vect/: Literal[Format[Vect[v_]]] := HoldForm[v]


(**************************** Result Functions *************************)
(*
Die vom Nutzer waehrend einer Sitzung erzeugten Symbole werden im
aktuellen Kontext (meist Global`) abgelegt. VASave speichert alle im
aktuellen Kontext befindlichen Definitionen in einen File. Es ist deshalb
wuenschenswert, wenn alle Definitionen, die ein vom Nutzer erzeugtes
Symbol betreffen, in den aktuellen Kontext eingeordnet werden.

Aus dieser Forderung entsteht folgendes Problem:
Sei v ein vom Nutzer definiertes Vektorfeld. Will man fuer rot[v], div[v]
Werte definieren, so ist dies nicht durch Auf-Wert-Definitionen wie
          v/: rot[v] := wert1
          v/: div[v] := wert2
moeglich, da die FullFormen von rot[v], div[v] die Gestalt rot[Vect[v]],
div[Vect[v]] haben. Es waeren nur Ab-Wert-Definitionen
          rot/: rot[v] := wert1
          div/: div[v] := wert2
zulaessig. Diese Definitionen gehoeren dann jedoch zu rot, div, die nicht
im aktuellen Kontext sondern im Kontext Vectan` liegen, der nicht sinnvoll
gesaved werden kann. (Er wuerde beim Laden zweimal geladen werden, einmal
mit dem Paket VECTAN, das zweite mal beim Laden der zusaetzlichen Nutzer-
definitionen. Danach waeren Originaldefinitionen mehrfach vorhanden, was
zu Strungen fhren kann.)

Zur Lsung dieses Problems werden Funktionen
          rotValue, divValue, gradValue, SPValue, VPValue
eingefuehrt. Diese tragen alle das Attribut HoldAll.
Die Definition eines Wertes fuer rot[v] kann nun erfolgen durch
          v/: rotValue[v] := wert . 
rot verwertet diese Information vermoege der Zeile        
          rot[Vect[x_]] := rotValue[x] /; ValueQ[rotValue[x]] .
Die anderen Funktionen werden analog genutzt. Bei Funktionen mit
mehreren Argumenten ordnet man die Definition einem der beiden 
Argumente zu. Beispiel:
          v/: SPValue[v,w] := 0
          SP[Vect[x_],Vect[y_]] := SPValue[x,y] /; ValueQ[SPValue[x,y]] .
*)

Attributes[gradValue] = {HoldAll}   (* Argumente: 1 oder 2 *)
Attributes[rotValue] = {HoldAll}    (* Argumente: 1        *)
Attributes[divValue] = {HoldAll}    (* Argumente: 1        *)
Attributes[SPValue] = {HoldAll}     (* Argumente: 2        *)
Attributes[VPValue] = {HoldAll}     (* Argumente: 2        *)

(*
Weitere solche Funktionen
            dotValue   und   graddotValue
werden in deriv.m definiert.
*)

(***************************** Scalar Product **************************)

Attributes[SP] = {Orderless} ;

Literal[SP[0,x_]] := 0 ;

Literal[SP[x_Plus,y_]] := SP[#,y]& /@ x ;

Literal[SP[Times[x_,y__],z_]] := x*SP[Times[y],z] /; !VectQ[x] ;

SP[f_,v_] := (Message[SP::error1,f];Abort[];) /; !VectQ[f] ;

Literal[SP[x_,VP[y_,z_]]] := Signature[{x,y,z}]*
                             Apply[ SP[#1,VP[#2,#3]]& , Sort[{x,y,z}] ] /; 
                                                  {x,y,z} =!=Sort[{x,y,z}]

Literal[SP[x_,VP[x_,y_]]] := 0 /; And[VectQ[x], VectQ[y]] ;

Literal[SP[VP[a_,b_],VP[x_,y_]]] := SP[a,x]*SP[b,y] - SP[b,x]*SP[a,y] ;

SP/: Literal[SP[a_,VP[b_,c_]]*SP[e_,VP[f_,g_]]] := 
                                     SP[a,e]*SP[b,f]*SP[c,g] +
                                     SP[a,f]*SP[b,g]*SP[c,e] +
                                     SP[a,g]*SP[b,e]*SP[c,f] -
                                     SP[a,g]*SP[b,f]*SP[c,e] -
                                     SP[a,e]*SP[b,g]*SP[c,f] - 
                                     SP[a,f]*SP[b,e]*SP[c,g] /; $SpatRule ;

SP/: Literal[Power[SP[a_,VP[b_,c_]],k_Integer]] :=
     Power[SP[a,VP[b,c]],k-2]*(2*SP[a,b]*SP[a,c]*SP[b,c] - (SP[a,c]^2)*SP[b,b]
                               - SP[a,a]*(SP[b,c]^2) - (SP[a,b]^2)*SP[c,c] +
                                 SP[a,a]*SP[b,b]*SP[c,c]) /; And[
                                                                $SpatRule,
                                                                2 <= k
                                                                ] ;

Literal[SP[Vect[x_],Vect[y_]]] := SPValue[x,y] /; ValueQ[SPValue[x,y]] ;
Literal[SP[Vect[x_],VP[Vect[y_],Vect[z_]]]] := SPValue[x,VP[y,z]] /;
                                               ValueQ[SPValue[x,VP[y,z]]];
Literal[SP[grad[f_],Vect[x_]]] := SPValue[grad[f],x] /;
                                  ValueQ[SPValue[grad[f],x]];
Literal[SP[rot[Vect[y_]],Vect[x_]]] := SPValue[rot[y],x] /;
                                       ValueQ[SPValue[rot[y],x]];
Literal[SP[grad[Vect[x_],Vect[y_]],Vect[z_]]] := SPValue[grad[x,y],z] /;
                                                 ValueQ[SPValue[grad[x,y],z]];


SP/: Literal[Format[SP[x_,y_]]] := SequenceForm[
                                               "<",
                                               HoldForm[x],
                                               ","
                                               HoldForm[y],
                                               ">"
                                               ] ;


(***************************** Vector Product **************************)

Literal[VP[0,x_]] := 0 ;
Literal[VP[x_,0]] := 0 ;
Literal[VP[x_,x_]] := 0 ;

Literal[VP[x_Plus,y_]] := VP[#,y]& /@ x ;
Literal[VP[x_,y_Plus]] := VP[x,#]& /@ y ;

Literal[VP[Times[x_,y__],z_]] := x*VP[Times[y],z] /; !VectQ[x] ;
Literal[VP[z_,Times[x_,y__]]] := x*VP[z,Times[y]] /; !VectQ[x] ;

Literal[VP[f_,v_]] := (Message[VP::error1,f];Abort[];) /; !VectQ[f] ;
Literal[VP[v_,f_]] := (Message[VP::error1,f];Abort[];) /; !VectQ[f] ;


Literal[VP[x_,y_]] := - VP[y,x] /; Sort[{x,y}] =!= {x,y} ;

(* Die folgende Regel setzt voraus, dass das innere VP stets als zweiter
Faktor auftritt. Dies ist deshalb immer erfuellt, da die einzigen
vektoriellen Groessen, die in VECTAN auftreten, die Koepfe
grad, rot, Vect, VP haben, so dass VP in der lexikographischen Ordnung
als letzter Kopf kommt. Die vorhergehende Regel vertauscht VP somit
stets auf die Stelle des zweiten Arguments. *)

Literal[VP[x_,VP[y_,z_]]] := SP[x,z]*y - SP[x,y]*z /; 
                                   And[$BacCab,VectQ[x],VectQ[y],VectQ[z]];

Literal[VP[Vect[x_],Vect[y_]]] := VPValue[x,y] /; ValueQ[VPValue[x,y]] ;
Literal[VP[grad[f_],Vect[x_]]] := VPValue[grad[f],x] /;
                                  ValueQ[VPValue[grad[f],x]] ;
Literal[VP[rot[Vect[x_]],Vect[y_]]] := VPValue[rot[y],x] /;
                                       ValueQ[VPValue[rot[y],x]];
Literal[VP[grad[Vect[x_],Vect[y_]],Vect[z_]]] := VPValue[grad[x,y],z] /;
                                                 ValueQ[VPValue[grad[x,y],z]];

VP/: Literal[Format[VP[x_,y_]]] := SequenceForm[
                                               "(",
                                               HoldForm[x],
                                               $VPChar,
                                               HoldForm[y],
                                               ")"
                                               ] ;


(******************************** Gradient *****************************)

Literal[grad[x_]] := 0 /; NumberQ[x];
Literal[grad[x_Symbol]] := 0 /; MemberQ[Attributes[x],Constant] ;

Literal[grad[x_Plus]] := grad /@ x;

Literal[grad[Times[f_,g__]]] := Times[grad[f],g] + f*grad[Times[g]];

Literal[grad[v_]] := (Message[grad::error1,v];Abort[];) /; VectQ[v] ;

Literal[grad[SP[v_,w_]]] := grad[v,w] + grad[w,v] + VP[v,rot[w]] +
                              VP[w,rot[v]] /; And[VectQ[v],VectQ[w]] ;


(*----------------------------- Kettenregel ---------------------------*)

Literal[grad[f_[g_]]] := f'[g]*grad[g] /; And[!MatchQ[f,Derivative[__]],
                                             !VectQ[g]
                                             ];

Literal[grad[f_[g__]]] :=
Chain[f,List[g],grad[#]&] /; And[!MatchQ[f,Derivative[__]],
                                And @@ (!VectQ[#]&) /@ List[g]
                                ];

Chain[f_, h_List, operator_] :=
(*
Fuehrt die Berechnung der Kettenregel fuer zusammengesetzte Funktionen
unter dem Differentialoperator operator aus. operator muss als reine
Funktion gegeben sein. f ist eine Funktion und h die Liste ihrer
Variablen.
*)
Module[{vars,i,func,derivs},
      vars = Table[
                  Slot[i],
                  {i,Length[h]}
                  ];
      func = Apply[f,vars];
      derivs = D[func,#]& /@ vars;
      derivs = Apply[(Evaluate[derivs]&),h];
      Expand[ derivs.(operator /@ h) ]
      ] ;
(*---------------------------------------------------------------------*)


Literal[grad[f_]] := gradValue[f] /; ValueQ[gradValue[f]] ;

(*----------------------- grad on vectors -----------------------------*)

Literal[grad[0,x_]] := 0 /; VectQ[x] ;
Literal[grad[x_,0]] := 0 /; VectQ[x] ;

Literal[grad[x_Plus,y_]] := grad[#,y]& /@ x;
Literal[grad[x_,y_Plus]] := grad[x,#]& /@ y;

Literal[grad[Times[f_,g__],h_]] := f*grad[Times[g],h] /; !VectQ[f] ;
Literal[grad[h_,Times[f_,g__]]] := Times[SP[grad[f],h],g] +
                                        f*grad[h,Times[g]] /; !VectQ[f] ;

Literal[grad[x_,VP[y_,z_]]] := VP[grad[x,y],z] + VP[y,grad[x,z]] /;
                                  And[VectQ[x],VectQ[y],VectQ[z]] ;

Literal[grad[Vect[x_],Vect[y_]]] := gradValue[x,y] /; ValueQ[gradValue[x,y]];
Literal[grad[Vect[x_],grad[f_]]] := gradValue[x,grad[f]] /; 
                                    ValueQ[gradValue[x,grad[f]]] ;
Literal[grad[Vect[x_],rot[Vect[y_]]]] := gradValue[x,rot[y]] /;
                                         ValueQ[gradValue[x,rot[y]]] ;
Literal[grad[Vect[x_],VP[Vect[y_],Vect[z_]]]] := gradValue[x,VP[y,z]] /;
                                                 ValueQ[gradValue[x,VP[y,z]]];
Literal[grad[Vect[x_],grad[Vect[y_],Vect[z_]]]] := gradValue[x,grad[y,z]] /;
                                             ValueQ[gradValue[x,grad[y,z]]];
Literal[grad[grad[f_],Vect[x_]]] := gradValue[grad[f],x] /;
                                    ValueQ[gradValue[grad[f],x]] ;
Literal[grad[rot[Vect[y_]],Vect[x_]]] := gradValue[rot[y],x] /;
                                         ValueQ[gradValue[rot[y],x]] ;
Literal[grad[VP[Vect[y_],Vect[z_]],Vect[x_]]] := gradValue[VP[y,z],x] /;
                                                 ValueQ[gradValue[VP[y,z],x]];
Literal[grad[grad[Vect[x_],Vect[y_]],Vect[z_]]] := gradValue[grad[x,y],z] /;
                                             ValueQ[gradValue[grad[x,y],z]];


grad/: Literal[Format[grad[v_,w_]]] := SequenceForm[
                                                   "(<",
                                                   HoldForm[v],
                                                   ",grad>)[",
                                                   HoldForm[w],
                                                   "]"
                                                   ] /;And[VectQ[v],VectQ[w]]


(******************************* Divergence ****************************)

Literal[div[0]] := 0 ;

Literal[div[x_Plus]] := div /@ x;

Literal[div[Times[f_,g__]]] := f*div[Times[g]] + SP[Times[g],grad[f]] /; 
                                                             !VectQ[f] ;

Literal[div[VP[v_,w_]]] := SP[w,rot[v]] - SP[v,rot[w]] /; 
                                 And[VectQ[v],VectQ[w]] ;

Literal[div[Times[f___,v_,w_]]] := 
       (Message[div::error1,v,w];Abort[];) /; And[VectQ[v],VectQ[w]] ;

Literal[div[f_]] := (Message[div::error2];Abort[];) /; !VectQ[f] ;

Literal[div[Vect[x_]]] := divValue[x] /; ValueQ[divValue[x]] ;


(******************************** Rotation *****************************)

Literal[rot[0]] := 0 ;

Literal[rot[x_Plus]] := rot /@ x;

Literal[rot[Times[f_, g__]]] := f*rot[Times[g]] + VP[grad[f],Times[g]] /; 
                                                              !VectQ[f] ;

Literal[rot[VP[v_,w_]]] := grad[w,v] - grad[v,w] + v*div[w] - w*div[v] /;
                                                 And[VectQ[v],VectQ[w]] ;

Literal[rot[Times[f___,v_,w_]]] := 
       (Message[rot::error1,v,w];Abort[];) /; And[VectQ[v],VectQ[w]] ;

Literal[rot[f_]] := (Message[rot::error2];Abort[];) /; !VectQ[f] ;

Literal[rot[Vect[x_]]] := rotValue[x] /; ValueQ[rotValue[x]] ;


(********************************** Rules ******************************)

Literal[rot[grad[v_]]] := 0 ;

Literal[div[rot[v_]]] := 0 ;

(********************************** BacCab *****************************)

BacCab[expr_,x_,y_,z_] :=
(*
Wendet auf den Ausdruck expr mittels ReplaceAll die BacCab-Regel
               VP[x,VP[y,z]] -> SP[x,z]y - SP[x,y]z
an.
*)
Module[{ls,rs},
      ls = VP[x,VP[y,z]];
      rs = SP[x,z]*y - SP[x,y]*z;
      expr /. {ls -> rs} /. {- ls -> - rs}
      ];

(******************************** KillSpat *****************************)

(*
SP/: Literal[Times[v___,SP[x_,VP[y_,z_]],Vect[u_]]] :=
            Times[v,
                 SP[z,Vect[u]]*VP[x,y] -
                 SP[y,Vect[u]]*VP[x,z] +
                 SP[x,Vect[u]]*VP[y,z]
                 ] /; And[$KillSpat,VectQ[x],VectQ[y],VectQ[z]];
*)

(*
SP/: Literal[ Times[w___,SP[x_,VP[y_,z_]],VP[Vect[u_],Vect[v_]]] ] :=
     Times[w,
          (SP[x,Vect[v]]*SP[Vect[u],z] - SP[x,Vect[u]]*SP[Vect[v],z])*y +
          (SP[x,Vect[u]]*SP[Vect[v],y] - SP[x,Vect[v]]*SP[Vect[u],y])*z +
          (SP[y,Vect[u]]*SP[Vect[v],z] - SP[z,Vect[u]]*SP[Vect[v],y])*x
          ] /; And[$KillSpat,VectQ[x],VectQ[y],VectQ[z]];
*)

SP/: Literal[ Times[w___,SP[x_,VP[y_,z_]],VP[Vect[u_],Vect[v_]]] ] :=
     Times[w,
          (SP[x,Vect[v]]*SP[Vect[u],z] - SP[x,Vect[u]]*SP[Vect[v],z])*y +
          (SP[x,Vect[u]]*SP[Vect[v],y] - SP[x,Vect[v]]*SP[Vect[u],y])*z +
          (SP[y,Vect[u]]*SP[Vect[v],z] - SP[z,Vect[u]]*SP[Vect[v],y])*x
          ] /; $KillSpat;

SP/: Literal[Times[v___,SP[x_,VP[y_,z_]],u_]] :=
            Times[v,
                 SP[z,u]*VP[x,y] -
                 SP[y,u]*VP[x,z] +
                 SP[x,u]*VP[y,z]
                 ] /; And[$KillSpat,VectQ[u]];

